@@ -18,6 +18,7 @@ from mcomix import constants
 from mcomix import file_provider
 from mcomix import callback
 from mcomix import log
+from mcomix import last_read_page
 
 class FileHandler(object):
 
@@ -55,18 +56,24 @@ class FileHandler(object):
         self._condition = None
         #: Provides a list of available files/archives in the open directory.
         self._file_provider = None
+        #: Keeps track of the last read page in archives
+        self.last_read_page = last_read_page.LastReadPage(constants.LASTPAGE_DATABASE_PATH)
         #: Regexp used for determining which archive files are images.
         self._image_re = constants.SUPPORTED_IMAGE_REGEX
         #: Regexp used for determining which archive files are comment files.
         self._comment_re = None
         self.update_comment_extensions()
 
+        self.last_read_page.set_enabled(
+            prefs['store recent file info'] == constants.STORE_LAST_PATH_AND_PAGE)
+
     def refresh_file(self, *args, **kwargs):
         """ Closes the current file(s)/archive and reloads them. """
-        current_file = os.path.abspath(self._window.imagehandler.get_real_path())
-        self.open_file(current_file, keep_fileprovider=True)
+        if self.file_loaded:
+            current_file = os.path.abspath(self._window.imagehandler.get_real_path())
+            self.open_file(current_file, keep_fileprovider=True)
 
-    def open_file(self, path, start_page=1, keep_fileprovider=False):
+    def open_file(self, path, start_page=0, keep_fileprovider=False):
         """Open the file pointed to by <path>.
 
         If <start_page> is not set we set the current
@@ -170,20 +177,21 @@ class FileHandler(object):
 
             # If no extraction is required, mark all files as available instantly.
             if self.archive_type is None:
-                for path in filelist:
-                    self.file_available(path)
+                self.file_available(filelist)
 
             result = True
 
         tools.alphanumeric_sort(self._comment_files)
 
         self._window.uimanager.recent.add(self._current_file)
-        self._window.draw_image()
+        # _window.draw_image is called as soon as the currently opened
+        # page is extracted/read via callback function.
 
         return result
 
     def close_file(self, *args):
         """Run tasks for "closing" the currently opened file(s)."""
+        self.update_last_read_page()
         self.file_loaded = False
         self.archive_type = None
         self._current_file = None
@@ -275,7 +283,7 @@ class FileHandler(object):
             self._condition = self._extractor.setup(self._base_path,
                                                 self._tmp_dir,
                                                 self.archive_type)
-        except Exception, ex:
+        except Exception:
             self._condition = None
             raise
 
@@ -310,16 +318,8 @@ class FileHandler(object):
                 self._name_table[full_path] = name
 
             # Determine current archive image index.
-            num_of_pages = len(image_files)
-            if start_page < 0:
-                if self._window.is_double_page:
-                    current_image_index = num_of_pages - 2
-                else:
-                    current_image_index = num_of_pages - 1
-            else:
-                current_image_index = start_page - 1
-
-            current_image_index = max(0, current_image_index)
+            current_image_index = self._get_index_for_page(start_page,
+                len(image_files), path)
 
             # Sort files to determine extraction order.
             self._sort_archive_files(archive_images, current_image_index)
@@ -380,15 +380,8 @@ class FileHandler(object):
                     image_files = tmp_image_files
 
             # Image index may have changed after additional files were extracted.
-            num_of_pages = len(image_files)
-            if start_page < 0:
-                if self._window.is_double_page:
-                    current_image_index = num_of_pages - 2
-                else:
-                    current_image_index = num_of_pages - 1
-            else:
-                current_image_index = start_page - 1
-            current_image_index = max(0, current_image_index)
+            current_image_index = self._get_index_for_page(start_page,
+                len(image_files), path)
 
             return image_files, current_image_index
 
@@ -413,6 +406,25 @@ class FileHandler(object):
         for key, value in state.iteritems():
             setattr(self, key, value)
 
+    def _get_index_for_page(self, start_page, num_of_pages, path):
+        """ Returns the page that should be displayed for an archive.
+        @param start_page: If -1, show last page. If 0, show either first page
+                           or last read page. If > 0, show C{start_page}.
+        @param num_of_pages: Page count.
+        @param path: Archive path.
+        """
+        if start_page < 0 and self._window.is_double_page:
+            current_image_index = num_of_pages - 2
+        elif start_page < 0 and not self._window.is_double_page:
+            current_image_index = num_of_pages - 1
+        elif start_page == 0:
+            # get_page returns either a page number (starting from 1) or None
+            current_image_index = (self.last_read_page.get_page(path) or 1) - 1
+        else:
+            current_image_index = start_page - 1
+
+        return min(max(0, current_image_index), num_of_pages - 1)
+
     def _sort_archive_files(self, archive_images, current_image_index):
         """ Sorts the list C{archive_images} in place based on a priority order
         algorithm. """
@@ -461,6 +473,9 @@ class FileHandler(object):
             self._condition.notifyAll()
             self._condition.release()
 
+        self.update_last_read_page()
+        self.last_read_page.cleanup()
+
     def get_number_of_comments(self):
         """Return the number of comments in the current archive."""
         return len(self._comment_files)
@@ -627,9 +642,10 @@ class FileHandler(object):
             return False
 
     @callback.Callback
-    def file_available(self, filepath):
+    def file_available(self, filepaths):
         """ Called every time a new file from the Filehandler's opened
-        files becomes available. <filepath> is the now available file. """
+        files becomes available. C{filepaths} is a list of now available files.
+        """
         pass
 
     def _extracted_file(self, name):
@@ -637,7 +653,7 @@ class FileHandler(object):
         <name>. This name is relative to the temporary directory
         the files were extracted to. """
         filepath = os.path.join(self._tmp_dir, name)
-        self.file_available(filepath)
+        self.file_available([filepath])
 
     def _wait_on_page(self, page):
         """Block the running (main) thread until the file corresponding to
@@ -716,5 +732,20 @@ class FileHandler(object):
 
         return fileinfo
 
+    def update_last_read_page(self):
+        """ Stores the currently viewed page. """
+        if self.archive_type is None or not self.file_loaded:
+            return
+
+        archive_path = self.get_path_to_base()
+        page = self._window.imagehandler.get_current_page()
+        is_last_page = self._window.imagehandler.is_last_page()
+        # Do not store first and last page (last page means archive was finished,
+        # first page is default behaviour and would waste space unnecessarily)
+        if page == 1 or is_last_page:
+            self.last_read_page.clear_page(archive_path)
+        else:
+            self.last_read_page.set_page(archive_path, page)
+
 
 # vim: expandtab:sw=4:ts=4
