@@ -27,18 +27,31 @@ class ImageHandler:
 
     def __init__(self, window):
 
-        self.first_wanted = 0
-        self.last_wanted = 1
-        self._current_file = None
-        self.is_cacheing = False
-
+        #: Reference to main window
         self._window = window
-        self._base_path = None
+
+        #: First index of last pixbuf cache run
+        self.first_wanted = 0
+        #: Last index of last pixbuf cache run
+        self.last_wanted = 1
+
+        #: Pixbufs are currently being cached
+        self.is_cacheing = False
+        #: Interrupts the cacheing thread if True
         self._stop_cacheing = False
+        #: Caching thread
+        self._thread = None
+
+        #: Archive path, if currently opened file is archive
+        self._base_path = None
+        #: List of image file names, either from extraction or directory
         self._image_files = None
+        #: Index of current page
         self._current_image_index = None
+        #: Pixbuf map from page > Pixbuf
         self._raw_pixbufs = {}
-        self._name_table = {}
+
+        #: Advance only one page instead of two in double page mode
         self.force_single_step = False
 
         self._window.filehandler.file_available += self._file_available
@@ -159,9 +172,9 @@ class ImageHandler:
     def thread_cache_new_pixbufs(self, wanted_pixbufs):
         """Start threaded cache loading.
         """
-        cache_thread = threading.Thread(target=self.cache_new_pixbufs, args=(wanted_pixbufs,))
-        cache_thread.setDaemon(False)
-        cache_thread.start()
+        self._thread = threading.Thread(target=self.cache_new_pixbufs, args=(wanted_pixbufs,))
+        self._thread.setDaemon(False)
+        self._thread.start()
 
     def cache_new_pixbufs(self, wanted_pixbufs):
         """Cache new pixbufs if they are not already cached.
@@ -176,13 +189,11 @@ class ImageHandler:
         self.is_cacheing = False
 
     def next_page(self):
-        """Set up filehandler to the next page. Return True if this results
-        in a new page.
+        """Set up filehandler to the next page. Return the new page number.
         """
         if not self._window.filehandler.file_loaded and self._window.filehandler.archive_type is None:
             return False
 
-        old_page = self.get_current_page()
         viewed = self._window.displayed_double() and 2 or 1
 
         if self.get_current_page() + viewed > self.get_number_of_pages():
@@ -199,7 +210,7 @@ class ImageHandler:
             if not next_archive_opened and \
                 prefs['auto open next directory'] and \
                 (not archive_open or prefs['auto open next archive']):
-                directory_changed = self._window.filehandler.open_next_directory()
+                self._window.filehandler.open_next_directory()
 
             return False
 
@@ -208,8 +219,7 @@ class ImageHandler:
         return self.get_current_page()
 
     def previous_page(self):
-        """Set up filehandler to the previous page. Return True if this
-        results in a new page.
+        """Set up filehandler to the previous page. Return the new page number.
         """
         if not self._window.filehandler.file_loaded and self._window.filehandler.archive_type is None:
             return False
@@ -228,11 +238,10 @@ class ImageHandler:
             if not previous_archive_opened and \
                 prefs['auto open next directory'] and \
                 (not archive_open or prefs['auto open next archive']):
-                directory_changed = self._window.filehandler.open_previous_directory()
+                self._window.filehandler.open_previous_directory()
 
             return False
 
-        old_page = self.get_current_page()
         step = self._get_backward_step_length()
         step = min(self._current_image_index, step)
         self._current_image_index -= step
@@ -243,22 +252,18 @@ class ImageHandler:
         return self.get_current_page()
 
     def first_page(self):
-        """Set up filehandler to the first page. Return True if this
-        results in a new page.
+        """Set up filehandler to the first page. Return the new page number.
         """
         if not self._window.filehandler.file_loaded:
             return False
-        old_page = self.get_current_page()
         self._current_image_index = 0
         return self.get_current_page()
 
     def last_page(self):
-        """Set up filehandler to the last page. Return True if this results
-        in a new page.
+        """Set up filehandler to the last page. Return the new page number.
         """
         if not self._window.filehandler.file_loaded:
             return False
-        old_page = self.get_current_page()
         offset = self._window.is_double_page and 2 or 1
         offset = min(self.get_number_of_pages(), offset)
         self._current_image_index = self.get_number_of_pages() - offset
@@ -267,13 +272,11 @@ class ImageHandler:
         return self.get_current_page()
 
     def set_page(self, page_num):
-        """Set up filehandler to the page <page_num>. Return True if this
-        results in a new page.
+        """Set up filehandler to the page <page_num>. Return the new page number.
         """
         if not 0 < page_num <= self.get_number_of_pages():
             return False
 
-        old_page = self.get_current_page()
         self._current_image_index = page_num - 1
         self.do_cacheing()
 
@@ -318,21 +321,23 @@ class ImageHandler:
 
         self.first_wanted = 0
         self.last_wanted = 1
-        self._current_file = None
 
+        self.cleanup()
         self._base_path = None
-        self._stop_cacheing = False
         self._image_files = []
         self._current_image_index = None
         self._raw_pixbufs.clear()
-        self._name_table.clear()
-        self.is_cacheing = False
 
         tools.garbage_collect()
 
     def cleanup(self):
         """Run clean-up tasks. Should be called prior to exit."""
         self._stop_cacheing = True
+        if self._thread:
+            self._thread.join()
+            self._thread = None
+        self._stop_cacheing = False
+        self.is_cacheing = False
 
     def page_is_available(self, page=None):
         """ Returns True if <page> is available and calls to get_pixbufs
@@ -360,16 +365,16 @@ class ImageHandler:
         file has been extracted. """
         pass
 
-    def _file_available(self, filepath):
+    def _file_available(self, filepaths):
         """ Called by the filehandler when a new file becomes available. """
         # Find the page that corresponds to <filepath>
         if not self._image_files:
             return
 
+        available = sorted(filepaths)
         for i, imgpath in enumerate(self._image_files):
-            if filepath == imgpath:
+            if tools.bin_search(available, imgpath) > -1:
                 self.page_available(i + 1)
-                break
 
     def is_last_page(self):
         """Return True if at the last page."""
