@@ -1,11 +1,14 @@
 """library_backend.py - Comic book library backend using sqlite."""
 
 import os
+import datetime
 
 from mcomix import archive_tools
 from mcomix import constants
 from mcomix import thumbnail_tools
 from mcomix import log
+from mcomix import callback
+from mcomix.library import backend_types
 
 try:
     from sqlite3 import dbapi2
@@ -16,14 +19,16 @@ except ImportError:
         log.warning( _('! Could neither find pysqlite2 nor sqlite3.') )
         dbapi2 = None
 
-class LibraryBackend:
+
+class _LibraryBackend:
 
     """The LibraryBackend handles the storing and retrieval of library
     data to and from disk.
     """
 
     #: Current version of the library database structure.
-    DB_VERSION = 1
+    # See method _upgrade_database() for changes between versions.
+    DB_VERSION = 3
 
     def __init__(self):
 
@@ -35,11 +40,14 @@ class LibraryBackend:
                 return row[0]
             return row
 
-        self._con = dbapi2.connect(constants.LIBRARY_DATABASE_PATH)
+        self._con = dbapi2.connect(constants.LIBRARY_DATABASE_PATH,
+            check_same_thread=False)
         self._con.row_factory = row_factory
 
+        self.watchlist = backend_types._WatchList(self)
+
         version = self._library_version()
-        self._upgrade_database(version, LibraryBackend.DB_VERSION)
+        self._upgrade_database(version, _LibraryBackend.DB_VERSION)
 
     def get_books_in_collection(self, collection=None, filter_string=None):
         """Return a sequence with all the books in <collection>, or *ALL*
@@ -202,9 +210,34 @@ class LibraryBackend:
         collection exists. Names are unique, so at most one such collection
         can exist.
         """
-        cur = self._con.execute('''select id from Collection
+        cur = self._con.execute('''select id, name, supercollection
+            from collection
             where name = ?''', (name,))
-        return cur.fetchone()
+        result = cur.fetchone()
+        cur.close()
+        if result:
+            return backend_types._Collection(*result)
+        else:
+            return None
+
+    def get_collection_by_id(self, id):
+        """ Returns the collection with ID C{id}.
+        @param id: Integer value. May be C{-1} or C{None} for default collection.
+        @return: L{_Collection} if found, None otherwise.
+        """
+        if id is None or id == -1:
+            return backend_types.DefaultCollection
+        else:
+            cur = self._con.execute('''select id, name, supercollection
+                from collection
+                where id = ?''', (id,))
+            result = cur.fetchone()
+            cur.close()
+
+            if result:
+                return backend_types._Collection(*result)
+            else:
+                return None
 
     def get_supercollection(self, collection):
         """Return the supercollection of <collection>."""
@@ -230,23 +263,40 @@ class LibraryBackend:
         old = self._con.execute('''select id from Book
             where path = ?''', (path,)).fetchone()
         try:
+            cursor = self._con.cursor()
             if old is not None:
-                self._con.execute('''update Book set
+                cursor.execute('''update Book set
                     name = ?, pages = ?, format = ?, size = ?
                     where path = ?''', (name, pages, format, size, path))
+                book_id = old
             else:
-                self._con.execute('''insert into Book
+                cursor.execute('''insert into Book
                     (name, path, pages, format, size)
                     values (?, ?, ?, ?, ?)''',
                     (name, path, pages, format, size))
+                book_id = cursor.lastrowid
+
+                book = backend_types._Book(book_id, name, path, pages,
+                        format, size, datetime.datetime.now().isoformat())
+                self.book_added(book)
+
+            cursor.close()
+
+            if collection is not None:
+                self.add_book_to_collection(book_id, collection)
+
+            return True
         except dbapi2.Error:
             log.error( _('! Could not add book "%s" to the library'), path )
             return False
-        if collection is not None:
-            book = self._con.execute('''select id from Book
-                where path = ?''', (path,)).fetchone()
-            self.add_book_to_collection(book, collection)
-        return True
+
+    @callback.Callback
+    def book_added(self, book):
+        """ Event that triggers when a new book is successfully added to the
+        library.
+        @param book: L{_Book} instance of the newly added book.
+        """
+        pass
 
     def add_collection(self, name):
         """Add a new collection with <name> to the library. Return True
@@ -342,6 +392,8 @@ class LibraryBackend:
 
     def remove_collection(self, collection):
         """Remove the <collection> (sans books) from the library."""
+        self._con.execute('''update watchlist set collection = NULL
+            where collection = ?''', (collection,))
         self._con.execute('delete from Collection where id = ?', (collection,))
         self._con.execute('delete from Contain where collection = ?',
             (collection,))
@@ -353,11 +405,19 @@ class LibraryBackend:
         self._con.execute('''delete from Contain
             where book = ? and collection = ?''', (book, collection))
 
+    def execute(self, *args):
+        """ Passes C{args} directly to the C{execute} method of the SQL
+        connection. """
+        return self._con.execute(*args)
+
     def close(self):
         """Commit changes and close cleanly."""
         self._con.commit()
         self._con.close()
 
+        global _backend
+        _backend = None
+
     def _table_exists(self, table):
         """ Checks if C{table} exists in the database. """
         cursor = self._con.cursor()
@@ -399,6 +459,7 @@ class LibraryBackend:
         self._create_table_collection()
         self._create_table_contain()
         self._create_table_info()
+        self._create_table_watchlist()
 
     def _upgrade_database(self, from_version, to_version):
         """ Performs sequential upgrades to the database, bringing
@@ -412,19 +473,32 @@ class LibraryBackend:
 
         if from_version != to_version:
             upgrades = range(from_version, to_version)
+            log.info(_("Upgrading library database version from %(from)d to %(to)d."),
+                { "from" : from_version, "to" : to_version })
 
             if 0 in upgrades:
                 # Upgrade from Comix database structure to DB version 1
                 # (Added table 'info')
-                log.info(_("Upgrading library database version from %(from)d to %(to)d."),
-                    { "from" : from_version, "to" : to_version })
                 self._create_table_info()
 
             if 1 in upgrades:
                 # Upgrade to database structure version 2.
-                # (No changes yet, but for me to remember that this if statement
-                # must not be 'elif' to preserve sequential upgrade bahaviour.)
-                pass
+                # (Added table 'watchlist' for storing auto-add directories)
+                self._create_table_watchlist()
+
+            if 2 in upgrades:
+                # Changed 'added' field in 'book' from date to datetime.
+                self._con.execute('''alter table book rename to book_old''')
+                self._create_table_book()
+                self._con.execute('''insert into book
+                    (id, name, path, pages, format, size, added)
+                    select id, name, path, pages, format, size, datetime(added)
+                    from book_old''')
+                self._con.execute('''drop table book_old''')
+
+
+            self._con.execute('''update info set value = ? where key = 'version' ''',
+                              (str(_LibraryBackend.DB_VERSION),))
 
     def _create_table_book(self):
         self._con.execute('''create table if not exists book (
@@ -434,7 +508,7 @@ class LibraryBackend:
             pages integer,
             format integer,
             size integer,
-            added date default current_date)''')
+            added datetime default current_timestamp)''')
 
     def _create_table_collection(self):
         self._con.execute('''create table if not exists collection (
@@ -454,7 +528,24 @@ class LibraryBackend:
             value text)''')
         self._con.execute('''insert into info
             (key, value) values ('version', ?)''',
-            (str(LibraryBackend.DB_VERSION),))
+            (str(_LibraryBackend.DB_VERSION),))
 
+    def _create_table_watchlist(self):
+        self._con.execute('''create table if not exists watchlist (
+            path text primary key,
+            collection integer references collection (id) on delete set null)''')
+
+
+_backend = None
+
+
+def LibraryBackend():
+    """ Returns the singleton instance of the library backend. """
+    global _backend
+    if _backend is not None:
+        return _backend
+    else:
+        _backend = _LibraryBackend()
+        return _backend
 
 # vim: expandtab:sw=4:ts=4
