@@ -2,22 +2,20 @@
 
 import os
 import urllib
-import Queue
-import threading
-import itertools
 import gtk
 import gobject
 import Image
 import ImageDraw
 
 from mcomix.preferences import prefs
+from mcomix import thumbnail_view
 from mcomix import file_chooser_library_dialog
 from mcomix import image_tools
 from mcomix import constants
 from mcomix import portability
-from mcomix import callback
 from mcomix import i18n
 from mcomix import status
+from mcomix import log
 from mcomix.library.pixbuf_cache import get_pixbuf_cache
 
 _dialog = None
@@ -26,6 +24,7 @@ _dialog = None
 # library's TreeModels.
 _COLLECTION_ALL = -1
 
+
 class _BookArea(gtk.ScrolledWindow):
 
     """The _BookArea is the central area in the library where the book
@@ -37,19 +36,29 @@ class _BookArea(gtk.ScrolledWindow):
 
         self._library = library
         self._cache = get_pixbuf_cache()
-        self._stop_update = False
-        self._thumbnail_threads = None
+
+        self._library.backend.book_added += self._new_book_added
 
         self.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
 
-        # Store Cover, book ID, book path, book size
+        # Store Cover, book ID, book path, book size, date added to library,
+        # is thumbnail loaded?
+
         # The SORT_ constants must correspond to the correct column here,
         # i.e. SORT_SIZE must be 3, since 3 is the size column in the ListStore.
         self._liststore = gtk.ListStore(gtk.gdk.Pixbuf,
-                gobject.TYPE_INT, gobject.TYPE_STRING, gobject.TYPE_INT64)
+                gobject.TYPE_INT, gobject.TYPE_STRING, gobject.TYPE_INT64,
+                gobject.TYPE_STRING, gobject.TYPE_BOOLEAN)
         self._liststore.set_sort_func(constants.SORT_NAME, self._sort_by_name, None)
-        self._iconview = gtk.IconView(self._liststore)
+        self.set_sort_order()
+        self._liststore.connect('row-inserted', self._icon_added)
+        self._iconview = thumbnail_view.ThumbnailIconView(self._liststore)
         self._iconview.set_pixbuf_column(0)
+        self._iconview.pixbuf_column = 0
+        self._iconview.status_column = 5
+        self._iconview.generate_thumbnail = self._get_pixbuf
+        self._iconview.get_file_path_from_model = lambda model, iter: \
+                model.get_value(iter, 2).decode('utf-8')
         self._iconview.connect('item_activated', self._book_activated)
         self._iconview.connect('selection_changed', self._selection_changed)
         self._iconview.connect_after('drag_begin', self._drag_begin)
@@ -58,7 +67,7 @@ class _BookArea(gtk.ScrolledWindow):
         self._iconview.connect('button_press_event', self._button_press)
         self._iconview.connect('key_press_event', self._key_press)
         self._iconview.connect('popup_menu', self._popup_menu)
-        self._iconview.modify_base(gtk.STATE_NORMAL, gtk.gdk.Color()) # Black.
+        self._iconview.modify_base(gtk.STATE_NORMAL, gtk.gdk.Color())  # Black.
         self._iconview.enable_model_drag_source(0,
             [('book', gtk.TARGET_SAME_APP, constants.LIBRARY_DRAG_EXTERNAL_ID)],
             gtk.gdk.ACTION_MOVE)
@@ -86,10 +95,13 @@ class _BookArea(gtk.ScrolledWindow):
                 <menuitem action="remove from library" />
                 <menuitem action="completely remove" />
                 <separator />
+                <menuitem action="copy to clipboard" />
+                <separator />
                 <menu action="sort">
                     <menuitem action="by name" />
                     <menuitem action="by path" />
                     <menuitem action="by size" />
+                    <menuitem action="by date added" />
                     <separator />
                     <menuitem action="ascending" />
                     <menuitem action="descending" />
@@ -135,6 +147,10 @@ class _BookArea(gtk.ScrolledWindow):
                 _('_Remove and delete from disk'), None,
                 _('Deletes the selected books from disk.'),
                 self._completely_remove_book),
+            ('copy to clipboard', gtk.STOCK_COPY,
+                _('_Copy'), None,
+                _('Copies the selected book\'s path to clipboard.'),
+                self._copy_selected),
             ('sort', None, _('_Sort'), None,
                 _('Changes the sort order of the library.'), None),
             ('cover size', None, _('Cover si_ze'), None,
@@ -144,13 +160,14 @@ class _BookArea(gtk.ScrolledWindow):
         actiongroup.add_radio_actions([
             ('by name', None, _('Book name'), None, None, constants.SORT_NAME),
             ('by path', None, _('Full path'), None, None, constants.SORT_PATH),
-            ('by size', None, _('File size'), None, None, constants.SORT_SIZE)],
+            ('by size', None, _('File size'), None, None, constants.SORT_SIZE),
+            ('by date added', None, _('Date added'), None, None, constants.SORT_LAST_MODIFIED)],
             prefs['lib sort key'], self._sort_changed)
         actiongroup.add_radio_actions([
             ('ascending', gtk.STOCK_SORT_ASCENDING, _('Ascending'), None, None,
-                constants.RESPONSE_SORT_ASCENDING),
+                constants.SORT_ASCENDING),
             ('descending', gtk.STOCK_SORT_DESCENDING, _('Descending'), None, None,
-                constants.RESPONSE_SORT_DESCENDING)],
+                constants.SORT_DESCENDING)],
             prefs['lib sort order'], self._sort_changed)
 
         # Library cover size
@@ -188,76 +205,49 @@ class _BookArea(gtk.ScrolledWindow):
         # order to not leak memory.
         self._liststore.clear()
 
-    def display_covers(self, collection):
-        """Display the books in <collection> in the IconView."""
+    def display_covers(self, collection_id):
+        """Display the books in <collection_id> in the IconView."""
 
         adjustment = self.get_vadjustment()
         if adjustment:
             adjustment.set_value(0)
 
         self.stop_update()
+        # Temporarily detach model to speed up updates
+        self._iconview.set_model(None)
         self._liststore.clear()
 
-        if collection == _COLLECTION_ALL: # The "All" collection is virtual.
-            collection = None
+        collection = self._library.backend.get_collection_by_id(collection_id)
+        books = collection.get_books(self._library.filter_string)
+        self.add_books(books)
 
-        # Get all books that need to be added.
-        # This cannot be executed threaded due to SQLite connections
-        # being bound to the thread that created them.
-        books = self._library.backend.get_books_in_collection(
-          collection, self._library.filter_string)
-        book_paths = [ self._library.backend.get_book_path(book)
-            for book in books ]
-        book_sizes = [ self._library.backend.get_book_size(book)
-            for book in books ]
-
-        filler = self._get_empty_thumbnail()
-        for book, path, size in itertools.izip(books, book_paths, book_sizes):
-
-            # Fill the liststore with a filler pixbuf.
-            iter = self._liststore.append([filler, book, path, size])
-
-        # Sort the list store based on preferences.
-        if prefs['lib sort order'] == constants.RESPONSE_SORT_ASCENDING:
-            sortorder = gtk.SORT_ASCENDING
-        else:
-            sortorder = gtk.SORT_DESCENDING
-        self.sort_books(prefs['lib sort key'], sortorder)
-
-        # Queue thumbnail loading
-        book_queue = Queue.Queue()
-        iter = self._liststore.get_iter_first()
-        while iter:
-            path = self._liststore.get_value(iter, 2)
-            wrapper = IterWrapper(iter, self._liststore)
-            self._path_about_to_be_removed += wrapper.invalidate
-            book_queue.put((wrapper, path.decode('utf-8')))
-
-            iter = self._liststore.iter_next(iter)
-
-        # Start the thumbnail threads.
-        self._thumbnail_threads = [ threading.Thread(target=self._pixbuf_worker,
-            args=(book_queue,)) for _ in range(3) ]
-        for thread in self._thumbnail_threads:
-            thread.setDaemon(True)
-            thread.start()
+        # Re-attach model here
+        self._iconview.set_model(self._liststore)
 
     def stop_update(self):
         """Signal that the updating of book covers should stop."""
-        self._stop_update = True
+        self._iconview.stop_update()
 
-        if self._thumbnail_threads:
-            for thread in self._thumbnail_threads:
-                thread.join()
+    def add_books(self, books):
+        """ Adds new book covers to the icon view.
+        @param books: List of L{_Book} instances. """
+        filler = self._get_empty_thumbnail()
 
-        # All update threads should have finished now.
-        self._stop_update = False
+        for book in books:
+            # Fill the liststore with a filler pixbuf.
+            self._liststore.append([filler, book.id,
+                book.path.encode('utf-8'), book.size, book.added, False])
+
+        self._iconview.draw_thumbnails_on_screen()
+
+    def _new_book_added(self, book):
+        """ Callback function for L{LibraryBackend.book_added}. """
+        self.add_books([book])
 
     def remove_book_at_path(self, path):
         """Remove the book at <path> from the ListStore (and thus from
         the _BookArea).
         """
-        self._path_about_to_be_removed(path)
         iterator = self._liststore.get_iter(path)
         filepath = self._liststore.get_value(iterator, 2)
         self._liststore.remove(iterator)
@@ -287,11 +277,16 @@ class _BookArea(gtk.ScrolledWindow):
             return
         self._book_activated(self._iconview, selected, True)
 
-    def sort_books(self, sort_key, sort_order=gtk.SORT_ASCENDING):
+    def set_sort_order(self):
         """ Orders the list store based on the key passed in C{sort_key}.
         Should be one of the C{SORT_} constants from L{constants}.
         """
-        self._liststore.set_sort_column_id(sort_key, sort_order)
+        if prefs['lib sort order'] == constants.SORT_ASCENDING:
+            sortorder = gtk.SORT_ASCENDING
+        else:
+            sortorder = gtk.SORT_DESCENDING
+
+        self._liststore.set_sort_column_id(prefs['lib sort key'], sortorder)
 
     def _sort_changed(self, old, current):
         """ Called whenever the sorting options changed. """
@@ -302,18 +297,15 @@ class _BookArea(gtk.ScrolledWindow):
             prefs['lib sort key'] = constants.SORT_PATH
         elif name == 'by size':
             prefs['lib sort key'] = constants.SORT_SIZE
+        elif name == 'by date added':
+            prefs['lib sort key'] = constants.SORT_LAST_MODIFIED
 
         if name == 'ascending':
-            prefs['lib sort order'] = constants.RESPONSE_SORT_ASCENDING
+            prefs['lib sort order'] = constants.SORT_ASCENDING
         elif name == 'descending':
-            prefs['lib sort order'] = constants.RESPONSE_SORT_DESCENDING
+            prefs['lib sort order'] = constants.SORT_DESCENDING
 
-        if prefs['lib sort order'] == constants.RESPONSE_SORT_ASCENDING:
-            order = gtk.SORT_ASCENDING
-        else:
-            order = gtk.SORT_DESCENDING
-
-        self.sort_books(prefs['lib sort key'], order)
+        self.set_sort_order()
 
     def _sort_by_name(self, treemodel, iter1, iter2, user_data):
         """ Compares two books based on their file name without the
@@ -327,8 +319,8 @@ class _BookArea(gtk.ScrolledWindow):
         elif path2 is None:
             return -1
 
-        name1 = os.path.split(path1)[1].lower()
-        name2 = os.path.split(path2)[1].lower()
+        name1 = os.path.split(path1.decode('utf-8'))[1].lower()
+        name2 = os.path.split(path2.decode('utf-8'))[1].lower()
 
         if name1 == name2:
             return 0
@@ -338,6 +330,14 @@ class _BookArea(gtk.ScrolledWindow):
             else:
                 return 1
 
+    def _icon_added(self, model, path, iter, *args):
+        """ Justifies the alignment of all cell renderers when new data is
+        added to the model. """
+        # FIXME Setting the alignment for all cells each time something
+        # is added seems rather wasteful. Find better way to do this.
+        for cell in self._iconview.get_cells():
+            cell.set_alignment(0.5, 0.5)
+
     def _book_size_changed(self, old, current):
         """ Called when library cover size changes. """
         old_size = prefs['library cover size']
@@ -386,41 +386,7 @@ class _BookArea(gtk.ScrolledWindow):
             collection = self._library.collection_area.get_current_collection()
             gobject.idle_add(self.display_covers, collection)
 
-    def _add_book(self, book):
-        """Add the <book> to the ListStore (and thus to the _BookArea)."""
-        path = self._library.backend.get_book_path(book)
-
-        if path:
-            pixbuf = self._get_pixbuf(path)
-            self._liststore.append([pixbuf, book])
-
-    def _pixbuf_worker(self, books):
-        """ Run by a worker thread to generate the thumbnail for a list
-        of books. """
-        while not self._stop_update and not books.empty():
-            try:
-                iter, path = books.get_nowait()
-            except Queue.Empty:
-                break
-
-            pixbuf = self._get_pixbuf(path)
-            gobject.idle_add(self._pixbuf_finished, (iter, pixbuf))
-            books.task_done()
-
-    def _pixbuf_finished(self, pixbuf_info):
-        """ Executed when a pixbuf was created, to actually insert the pixbuf
-        into the view store. <pixbuf_info> is a tuple containing (index, pixbuf). """
-
-        iterwrapper, pixbuf = pixbuf_info
-        iter = iterwrapper.iter
-
-        if iter and self._liststore.iter_is_valid(iter):
-            self._liststore.set(iter, 0, pixbuf)
-
-        # Remove this idle handler.
-        return 0
-
-    def _get_pixbuf(self, path):
+    def _get_pixbuf(self, path, model, model_path):
         """ Get or create the thumbnail for the selected book at <path>. """
         if self._cache.exists(path):
             return self._cache.get(path)
@@ -475,9 +441,12 @@ class _BookArea(gtk.ScrolledWindow):
             self._library.backend.remove_book_from_collection(book, collection)
             self.remove_book_at_path(path)
         coll_name = self._library.backend.get_collection_name(collection)
+        message = i18n.get_translation().ungettext(
+                "Removed %(num)d book from '%(collection)s'.",
+                "Removed %(num)d books from '%(collection)s'.",
+                len(selected))
         self._library.set_status_message(
-            _("Removed %(num)d book(s) from '%(collection)s'.") %
-            {'num': len(selected), 'collection': coll_name})
+            message % {'num': len(selected), 'collection': coll_name})
 
     def _remove_books_from_library(self, *args):
         """Remove the currently selected books from the library, and thus
@@ -491,7 +460,7 @@ class _BookArea(gtk.ScrolledWindow):
             self._library.backend.remove_book(book)
             self.remove_book_at_path(path)
 
-        msg = i18n.get_translation().ngettext(
+        msg = i18n.get_translation().ungettext(
             'Removed %d book from the library.',
             'Removed %d books from the library.',
             len(selected))
@@ -537,7 +506,18 @@ class _BookArea(gtk.ScrolledWindow):
                     # of a single file
                     os.remove(book_path)
                 except Exception:
-                    print _('! Could not remove file "%s"') % book_path
+                    log.error(_('! Could not remove file "%s"') % book_path)
+
+    def _copy_selected(self, *args):
+        """ Copies the currently selected item to clipboard. """
+        paths = self._iconview.get_selected_items()
+        if len(paths) == 1:
+            model = self._iconview.get_model()
+            iter = model.get_iter(paths[0])
+            path = model.get_value(iter, 2).decode('utf-8')
+            pixbuf = model.get_value(iter, 0)
+
+            self._library._window.clipboard.copy(path, pixbuf)
 
     def _button_press(self, iconview, event):
         """Handle mouse button presses on the _BookArea."""
@@ -558,7 +538,8 @@ class _BookArea(gtk.ScrolledWindow):
     def _popup_book_menu(self):
         """ Shows the book panel popup menu. """
 
-        books_selected = len(self._iconview.get_selected_items()) > 0
+        selected = self._iconview.get_selected_items()
+        books_selected = len(selected) > 0
         collection = self._library.collection_area.get_current_collection()
         is_collection_all = collection == _COLLECTION_ALL
 
@@ -568,6 +549,7 @@ class _BookArea(gtk.ScrolledWindow):
         self._set_sensitive('_title', False)
         self._set_sensitive('add', collection is not None)
         self._set_sensitive('remove from collection', books_selected and not is_collection_all)
+        self._set_sensitive('copy to clipboard', len(selected) == 1)
 
         menu = self._ui_manager.get_widget('/library books')
         menu.popup(None, None, None, 3, gtk.get_current_event_time())
@@ -664,27 +646,5 @@ class _BookArea(gtk.ScrolledWindow):
         collection_name = self._library.backend.get_collection_name(collection)
         self._library.add_books(paths, collection_name)
 
-    @callback.Callback
-    def _path_about_to_be_removed(self, path):
-        """ This will be called before a liststore path is about to be removed.
-        Used to invalidate the IterWrapper before PyGTK gets a chance to access
-        the invalid iterator. """
-        pass
-
-class IterWrapper(object):
-    """ This is a security wrapper for TreeIterator that gets notified when an
-    item is about to be deleted from the TreeControl's list store. It then
-    invalidates its iterator to prevent PyGTK from causing a segmentation fault
-    by accessing the iterator. """
-
-    def __init__(self, iter, liststore):
-        self.iter = iter
-        self.liststore = liststore
-
-    def invalidate(self, path):
-        if self.iter:
-            iter_path = self.liststore.get_path(self.iter)
-            if iter_path == path:
-                self.iter = None
 
 # vim: expandtab:sw=4:ts=4
