@@ -25,6 +25,8 @@ from mcomix import thumbbar
 from mcomix import clipboard
 from mcomix import pageselect
 from mcomix import osd
+from mcomix import keybindings
+
 
 class MainWindow(gtk.Window):
 
@@ -44,14 +46,14 @@ class MainWindow(gtk.Window):
         self.is_fullscreen = False
         self.is_double_page = False
         self.is_manga_mode = False
-        self.is_virtual_double_page = False # I.e. a wide image is displayed
+        self.is_virtual_double_page = False  # I.e. a wide image is displayed
         self.width = None
         self.height = None
         self.was_out_of_focus = False
         #: Used to remember if changing to fullscreen enabled 'Hide all'
         self.hide_all_forced = False
 
-        self._manual_zoom = 100 # In percent of original image size
+        self._manual_zoom = 100  # In percent of original image size
         self._waiting_for_redraw = False
 
         self._image_box = gtk.HBox(False, 2)
@@ -90,6 +92,9 @@ class MainWindow(gtk.Window):
         self.set_size_request(300, 300)  # Avoid making the window *too* small
         self.resize(prefs['window width'], prefs['window height'])
 
+        # Hook up any additional key events not handled by UI accelerators
+        self._event_handler.register_key_events()
+
         # This is a hack to get the focus away from the toolbar so that
         # we don't activate it with space or some other key (alternative?)
         self.toolbar.set_focus_child(
@@ -268,7 +273,7 @@ class MainWindow(gtk.Window):
     def draw_image(self, at_bottom=False, scroll=False):
         """Draw the current pages and update the titlebar and statusbar.
         """
-        if not self._waiting_for_redraw: # Don't stack up redraws.
+        if not self._waiting_for_redraw:  # Don't stack up redraws.
             self._waiting_for_redraw = True
             gobject.idle_add(self._draw_image, at_bottom, scroll,
                 priority=gobject.PRIORITY_HIGH_IDLE)
@@ -276,9 +281,6 @@ class MainWindow(gtk.Window):
     def _draw_image(self, at_bottom, scroll):
         self._display_active_widgets()
 
-        while gtk.events_pending():
-            gtk.main_iteration(False)
-
         if not self.filehandler.file_loaded:
             self._waiting_for_redraw = False
             return False
@@ -335,7 +337,7 @@ class MainWindow(gtk.Window):
                     total_width += right_unscaled_x
                     total_height += right_unscaled_y
 
-                total_width += 2 # For the 2 px gap between images.
+                total_width += 2  # For the 2 px gap between images.
                 scaled_width = int(self._manual_zoom * total_width / 100)
                 scaled_height = int(self._manual_zoom * total_height / 100)
                 scale_up = True
@@ -474,9 +476,6 @@ class MainWindow(gtk.Window):
         self._update_page_information()
         self._waiting_for_redraw = False
 
-        while gtk.events_pending():
-            gtk.main_iteration(False)
-
         return False
 
     def _update_page_information(self):
@@ -510,7 +509,7 @@ class MainWindow(gtk.Window):
         """ Called whenever a new page is ready for displaying. """
         # Refresh display when currently opened page becomes available.
         if page == self.imagehandler.get_current_page() \
-            or self.displayed_double() and page == self.imagehandler.get_current_page() + 1:
+            or (self.displayed_double() and page == self.imagehandler.get_current_page() + 1):
 
             self.draw_image(False, True)
 
@@ -535,7 +534,6 @@ class MainWindow(gtk.Window):
         self.draw_image(at_bottom=at_bottom, scroll=True)
 
     def set_page(self, num, at_bottom=False):
-
         if self.imagehandler.set_page(num):
             self.new_page(at_bottom=at_bottom)
             self.slideshow.update_delay()
@@ -562,7 +560,7 @@ class MainWindow(gtk.Window):
             self.set_page(new_page_number)
 
     def page_select(self, *args):
-        page_selector = pageselect.Pageselector(self)
+        pageselect.Pageselector(self)
 
     def rotate_90(self, *args):
         prefs['rotation'] = (prefs['rotation'] + 90) % 360
@@ -671,7 +669,7 @@ class MainWindow(gtk.Window):
 
     def manual_zoom_in(self, *args):
         new_zoom = self._manual_zoom * 1.15
-        if 95 < new_zoom < 105: # To compensate for rounding errors
+        if 95 < new_zoom < 105:  # To compensate for rounding errors
             new_zoom = 100
         if new_zoom > 1000:
             return
@@ -680,7 +678,7 @@ class MainWindow(gtk.Window):
 
     def manual_zoom_out(self, *args):
         new_zoom = self._manual_zoom / 1.15
-        if 95 < new_zoom < 105: # To compensate for rounding errors
+        if 95 < new_zoom < 105:  # To compensate for rounding errors
             new_zoom = 100
         if new_zoom < 10:
             return
@@ -1081,6 +1079,10 @@ class MainWindow(gtk.Window):
         preferences.write_preferences_file()
         self.uimanager.bookmarks.write_bookmarks_file()
 
+        # Write keyboard accelerator map
+        gtk.accel_map_save(constants.KEYBINDINGS_PATH)
+        keybindings.keybinding_manager().save()
+
     def save_and_terminate_program(self, *args):
         prefs['previous quit was quit and save'] = True
 
@@ -1122,10 +1124,12 @@ class MainWindow(gtk.Window):
 #: Main window instance
 __main_window = None
 
+
 def main_window():
     """ Returns the global main window instance. """
     return __main_window
 
+
 def set_main_window(window):
     global __main_window
     __main_window = window
