@@ -22,7 +22,7 @@ class _PreferencesDialog(gtk.Dialog):
 
         reset = self.add_button(_('Clear dialog choices'), constants.RESPONSE_REVERT_TO_DEFAULT)
         reset.set_tooltip_text(_('Clears all dialog choices that you have previously chosen not to be asked again.'))
-        close = self.add_button(gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE)
+        self.add_button(gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE)
 
         self._window = window
         self.set_has_separator(False)
@@ -213,10 +213,11 @@ class _PreferencesDialog(gtk.Dialog):
         page.add_row(step_length_button)
 
         label = gtk.Label(_('Show only one page where appropriate:'))
-        label.set_tooltip_text(
+        doublepage_control = self._create_doublepage_as_one_control()
+        doublepage_control.set_tooltip_text(
             _("When showing the first page of an archive, or an image's width "
               "exceeds its height, only a single page will be displayed."))
-        page.add_row(label, self._create_doublepage_as_one_control())
+        page.add_row(label, doublepage_control)
         page.new_section(_('Files'))
 
         auto_open_last_button = gtk.CheckButton(
@@ -228,14 +229,12 @@ class _PreferencesDialog(gtk.Dialog):
             _('Automatically open, on startup, the file that was open when MComix was last closed.'))
         page.add_row(auto_open_last_button)
 
-        store_recent_button = gtk.CheckButton(
-            _('Store information about recently opened files'))
-        store_recent_button.set_active(prefs['store recent file info'])
-        store_recent_button.connect('toggled', self._check_button_cb,
-            'store recent file info')
-        store_recent_button.set_tooltip_text(
+        store_recent_label = gtk.Label(
+            _('Store information about recently opened files:'))
+        store_recent_box = self._create_store_recent_combobox()
+        store_recent_box.set_tooltip_text(
             _('Add information about all files opened from within MComix to the shared recent files list.'))
-        page.add_row(store_recent_button)
+        page.add_row(store_recent_label, store_recent_box)
 
         return page
 
@@ -312,6 +311,11 @@ class _PreferencesDialog(gtk.Dialog):
         language_box = self._create_language_control()
         page.add_row(label, language_box)
 
+        page.new_section(_('File order'))
+
+        label = gtk.Label(_('Order files by:'))
+        page.add_row(label, self._create_sort_by_control())
+
         page.new_section(_('Cache'))
 
         create_thumbs_button = gtk.CheckButton(
@@ -323,15 +327,6 @@ class _PreferencesDialog(gtk.Dialog):
             _('Store thumbnails for opened files according to the freedesktop.org specification. These thumbnails are shared by many other applications, such as most file managers.'))
         page.add_row(create_thumbs_button)
 
-        delay_thumbs_button = gtk.CheckButton(
-            _('Delay thumbnail generation'))
-        delay_thumbs_button.set_tooltip_text(
-            _('Thumbnails are generated only when required, instead of directly after opening a new file.'))
-        delay_thumbs_button.set_active(prefs['delay thumbnails'])
-        delay_thumbs_button.connect('toggled', self._check_button_cb,
-            'delay thumbnails')
-        page.add_row(delay_thumbs_button)
-
         label = gtk.Label(_('Maximum number of pages to store in the cache:'))
         adjustment = gtk.Adjustment(prefs['max pages to cache'], -1, 500, 1, 3)
         cache_spinner = gtk.SpinButton(adjustment, digits=0)
@@ -415,27 +410,8 @@ class _PreferencesDialog(gtk.Dialog):
             (_('Chinese (traditional)'), 'zh_TW')]
         languages.sort(key=operator.itemgetter(0))
 
-        model = gtk.ListStore(gobject.TYPE_STRING, gobject.TYPE_STRING)
-        for name, lang_code in languages:
-            model.append((name, lang_code))
-
-        box = gtk.ComboBox(model)
-
-        # Determine current language
-        iter = model.get_iter_first()
-        index = 0
-        while iter:
-            if model.get_value(iter, 1) == prefs['language']:
-                box.set_active(index)
-                break
-            else:
-                iter = model.iter_next(iter)
-                index += 1
-
-        cell = gtk.CellRendererText()
-        box.pack_start(cell, True)
-        box.add_attribute(cell, 'text', 0)
-        box.connect('changed', self._language_changed_cb)
+        box = self._create_combobox(languages, prefs['language'],
+                self._language_changed_cb)
 
         return box
 
@@ -449,42 +425,163 @@ class _PreferencesDialog(gtk.Dialog):
 
     def _create_doublepage_as_one_control(self):
         """ Creates the ComboBox control for selecting virtual double page options. """
-        options = (
+        items = (
                 (_('Never'), 0),
                 (_('Only for title pages'), constants.SHOW_DOUBLE_AS_ONE_TITLE),
                 (_('Only for wide images'), constants.SHOW_DOUBLE_AS_ONE_WIDE),
                 (_('Always'), constants.SHOW_DOUBLE_AS_ONE_TITLE | constants.SHOW_DOUBLE_AS_ONE_WIDE))
 
-        model = gtk.ListStore(gobject.TYPE_INT, gobject.TYPE_STRING)
-        for text, value in options:
-            model.append((value, text))
+        box = self._create_combobox(items,
+                prefs['virtual double page for fitting images'],
+                self._double_page_changed_cb)
 
-        virtual_double_box = gtk.ComboBox(model)
-        renderer = gtk.CellRendererText()
-        virtual_double_box.pack_start(renderer, True)
-        virtual_double_box.add_attribute(renderer, "text", 1)
-
-        # Set active box option
-        iter = model.get_iter_first()
-        while iter:
-            if model.get_value(iter, 0) == prefs['virtual double page for fitting images']:
-                virtual_double_box.set_active_iter(iter)
-                break
-            else:
-                iter = model.iter_next(iter)
-
-        virtual_double_box.connect('changed', self._double_page_changed_cb)
-
-        return virtual_double_box
+        return box
 
     def _double_page_changed_cb(self, combobox, *args):
         """ Called when a new option was selected for the virtual double page option. """
         iter = combobox.get_active_iter()
         if combobox.get_model().iter_is_valid(iter):
-            value = combobox.get_model().get_value(iter, 0)
+            value = combobox.get_model().get_value(iter, 1)
             prefs['virtual double page for fitting images'] = value
             self._window.draw_image()
 
+    def _create_sort_by_control(self):
+        """ Creates the ComboBox control for selecting archive sort by options. """
+        sortkey_items = (
+                (_('No sorting'), 0),
+                (_('File name'), constants.SORT_NAME),
+                (_('File size'), constants.SORT_SIZE),
+                (_('Last modified'), constants.SORT_LAST_MODIFIED))
+
+        sortkey_box = self._create_combobox(sortkey_items, prefs['sort by'],
+            self._sort_by_changed_cb)
+
+        sortorder_items = (
+                (_('Ascending'), constants.SORT_ASCENDING),
+                (_('Descending'), constants.SORT_DESCENDING))
+
+        sortorder_box = self._create_combobox(sortorder_items,
+                prefs['sort order'],
+                self._sort_order_changed_cb)
+
+        box = gtk.HBox()
+        box.pack_start(sortkey_box)
+        box.pack_start(sortorder_box)
+
+        label = _("Files will be opened and displayed according to the sort order "
+              "specified here. This option does not affect ordering within archives.")
+        sortkey_box.set_tooltip_text(label)
+        sortorder_box.set_tooltip_text(label)
+
+        return box
+
+    def _sort_by_changed_cb(self, combobox, *args):
+        """ Called when a new option was selected for the virtual double page option. """
+        iter = combobox.get_active_iter()
+        if combobox.get_model().iter_is_valid(iter):
+            value = combobox.get_model().get_value(iter, 1)
+            prefs['sort by'] = value
+
+            self._window.filehandler.refresh_file()
+
+    def _sort_order_changed_cb(self, combobox, *args):
+        """ Called when sort order changes (ascending or descending) """
+        iter = combobox.get_active_iter()
+        if combobox.get_model().iter_is_valid(iter):
+            value = combobox.get_model().get_value(iter, 1)
+            prefs['sort order'] = value
+
+            self._window.filehandler.refresh_file()
+
+    def _create_store_recent_combobox(self):
+        """ Creates the combobox for "Store recently opened files". """
+        items = (
+                (_('Never'), 0),
+                (_('Only file names'), constants.STORE_LAST_PATH),
+                (_('File names and last read page'), constants.STORE_LAST_PATH_AND_PAGE))
+
+        # Map legacy true/false values:
+        if prefs['store recent file info'] is True:
+            selection = constants.STORE_LAST_PATH
+        elif prefs['store recent file info'] is False:
+            selection = 0
+        else:
+            selection = prefs['store recent file info']
+
+        box = self._create_combobox(items, selection, self._store_recent_changed_cb)
+        return box
+
+    def _store_recent_changed_cb(self, combobox, *args):
+        """ Called when option "Store recently opened files" was changed. """
+        iter = combobox.get_active_iter()
+        if not combobox.get_model().iter_is_valid(iter):
+            return
+
+        value = combobox.get_model().get_value(iter, 1)
+        last_value = prefs['store recent file info']
+        prefs['store recent file info'] = value
+        self._window.filehandler.last_read_page.set_enabled(
+            value == constants.STORE_LAST_PATH_AND_PAGE)
+
+        # If "Never" was selected, ask to purge recent files.
+        if (last_value > 0 and value == 0
+            and (self._window.uimanager.recent.count() > 0
+                 or self._window.filehandler.last_read_page.count() > 0)):
+
+            dialog = gtk.MessageDialog(self, gtk.DIALOG_MODAL,
+                gtk.MESSAGE_INFO, gtk.BUTTONS_YES_NO)
+            dialog.set_default_response(gtk.RESPONSE_YES)
+            dialog.set_markup('<span weight="bold" size="larger">' +
+                _('Delete information about recently opened files?') +
+                '</span>')
+            dialog.format_secondary_text(
+                _('This will remove all entries from the "Recent" menu,'
+                  ' and clear information about last read pages.'))
+            response = dialog.run()
+            dialog.destroy()
+
+            if response == gtk.RESPONSE_YES:
+                self._window.uimanager.recent.remove_all()
+                self._window.filehandler.last_read_page.clear_all()
+
+    def _create_combobox(self, options, selected_value, change_callback):
+        """ Creates a new dropdown combobox and populates it with the items
+        passed in C{options}.
+
+        @param options: List of tuples: (Option display text, option value)
+        @param selected_value: One of the values passed in C{options} that will
+            be pre-selected when the control is created.
+        @param change_callback: Function that will be called when the 'changed'
+            event is triggered.
+        @returns gtk.ComboBox
+        """
+        assert options and len(options[0]) == 2, "Invalid format for options."
+
+        # Use the first list item to determine typing of model fields.
+        # First field is textual description, second field is value.
+        model = gtk.ListStore(gobject.TYPE_STRING, type(options[0][1]))
+        for text, value in options:
+            model.append((text, value))
+
+        box = gtk.ComboBox(model)
+        renderer = gtk.CellRendererText()
+        box.pack_start(renderer, True)
+        box.add_attribute(renderer, "text", 0)
+
+        # Set active box option
+        iter = model.get_iter_first()
+        while iter:
+            if model.get_value(iter, 1) == selected_value:
+                box.set_active_iter(iter)
+                break
+            else:
+                iter = model.iter_next(iter)
+
+        if change_callback:
+            box.connect('changed', change_callback)
+
+        return box
+
     def _check_button_cb(self, button, preference):
         """Callback for all checkbutton-type preferences."""
 
