@@ -1,6 +1,7 @@
 """ui.py - UI definitions for main window.
 """
 
+import os.path
 import gtk
 
 from mcomix import bookmark_menu
@@ -11,7 +12,6 @@ from mcomix import recent
 from mcomix import dialog_handler
 from mcomix import constants
 from mcomix import status
-from mcomix import file_chooser_base_dialog
 from mcomix import file_chooser_main_dialog
 from mcomix.library import main_dialog as library_main_dialog
 
@@ -365,6 +365,9 @@ class MainUI(gtk.UIManager):
         self.get_widget('/Popup/menu_recent').set_submenu(self.recentPopup)
         self.get_widget('/Popup/menu_recent').show()
 
+        # Load keyboard accelerator map
+        if os.path.isfile(constants.KEYBINDINGS_PATH):
+            gtk.accel_map_load(constants.KEYBINDINGS_PATH)
         window.add_accel_group(self.get_accel_group())
 
         # Is there no built-in way to do this?
